# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import unittest

import pytest

from blackhole.config import Config
from blackhole.smtp import Smtp

from ._utils import (Args, cleandir, create_config, create_file, reset)


@pytest.mark.usefixtures('reset', 'cleandir')
class TestHeadersSwitchDisabled(unittest.TestCase):

    def test_headers_disabled(self):
        cfile = create_config(('dynamic_switch=false', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.mode == 'accept'
        smtp.process_header('x-blackhole-mode: bounce')
        assert smtp.mode == 'accept'

    def test_headers_enabled(self):
        cfile = create_config(('dynamic_switch=true', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.mode == 'accept'
        smtp.process_header('x-blackhole-mode: bounce')
        assert smtp.mode == 'bounce'

    def test_headers_default(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.mode == 'accept'
        smtp.process_header('x-blackhole-mode: bounce')
        assert smtp.mode == 'bounce'


@pytest.mark.usefixtures('reset', 'cleandir')
class TestDynamicSwitchDisabledByFlags(unittest.TestCase):

    def test_mode(self):
        smtp = Smtp([])
        smtp.mode = 'accept'
        smtp._flags = {'mode': 'bounce'}
        assert smtp.mode == 'bounce'

    def test_delay(self):
        smtp = Smtp([])
        smtp.delay = '30'
        smtp._flags = {'delay': 20}
        assert smtp.delay == 20

    def test_delay_range(self):
        smtp = Smtp([])
        smtp.delay = '30'
        smtp._flags = {'delay': ['10', '20']}
        assert smtp.delay in (10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20)

    def test_delay_and_mode(self):
        smtp = Smtp([])
        smtp.delay = '30'
        smtp.mode = 'accept'
        smtp._flags = {'delay': '20', 'mode': 'bounce'}
        assert smtp.delay == 20
        assert smtp.mode == 'bounce'

    def test_delay_range_and_mode(self):
        smtp = Smtp([])
        smtp.delay = '30'
        smtp.mode = 'accept'
        smtp._flags = {'delay': ['10', '20'], 'mode': 'bounce'}
        assert smtp.delay in (10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20)
        assert smtp.mode == 'bounce'


@pytest.mark.usefixtures('reset', 'cleandir')
class TestProcessHeaders(unittest.TestCase):

    def test_valid_mode_header(self):
        smtp = Smtp([])
        assert smtp.mode == 'accept'
        smtp.process_header('x-blackhole-mode: bounce')
        assert smtp.mode == 'bounce'

    def test_invalid_mode_header(self):
        smtp = Smtp([])
        assert smtp.mode == 'accept'
        smtp.process_header('x-blackhole-mode: help')
        assert smtp.mode == 'accept'

    def test_invalid_mode_header2(self):
        smtp = Smtp([])
        assert smtp.mode == 'accept'
        smtp.process_header('x-some-mode: bounce')
        assert smtp.mode == 'accept'

    def test_valid_single_delay(self):
        smtp = Smtp([])
        assert smtp.delay is None
        smtp.process_header('x-blackhole-delay: 30')
        assert smtp.delay is 30

    def test_invalid_single_delay(self):
        smtp = Smtp([])
        assert smtp.delay is None
        smtp.process_header('x-blackhole-delay: abc')
        assert smtp.delay is None

    def test_valid_range_delay(self):
        smtp = Smtp([])
        assert smtp.delay is None
        smtp.process_header('x-blackhole-delay: 5, 10')
        assert smtp.delay in [5, 6, 7, 8, 9, 10]

    def test_invalid_range_delay(self):
        smtp = Smtp([])
        assert smtp.delay is None
        smtp.process_header('x-blackhole-delay: abc, def')
        assert smtp.delay is None


@pytest.mark.usefixtures('reset', 'cleandir')
class TestModeSwitch(unittest.TestCase):

    def test_mode_default(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.mode == 'accept'

    def test_mode_invalid(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.mode = 'kura'
        assert smtp.mode == 'accept'

    def test_mode_valid(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.mode = 'bounce'
        assert smtp.mode == 'bounce'

    def test_mode_valid_overrides_config(self):
        cfile = create_config(('mode=bounce', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.mode == 'bounce'
        smtp.mode = 'accept'
        assert smtp.mode == 'accept'


@pytest.mark.usefixtures('reset', 'cleandir')
class TestDelaySwitch(unittest.TestCase):

    def test_delay_not_enabled_or_set(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.delay is None

    def test_delay_from_config(self):
        cfile = create_config(('delay=30', ))
        Config(cfile).load()
        smtp = Smtp([])
        assert smtp.delay is 30

    def test_delay_switch_overrides_config_single(self):
        cfile = create_config(('delay=30', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '60'
        assert smtp.delay is 60
        assert smtp.config.delay is 30

    def test_delay_switch_range_overrides_config(self):
        cfile = create_config(('delay=30', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '40, 45'
        assert smtp.delay in [x for x in range(40, 46)]
        assert smtp.config.delay is 30

    def test_delay_switch_invalid_single_value_no_config(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = 'fifteen'
        assert smtp.delay is None

    def test_delay_switch_invalid_single_value_config_60(self):
        cfile = create_config(('delay=30', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = 'fifteen'
        assert smtp.delay is 30

    def test_delay_switch_invalid_single_negative_value(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '-10'
        assert smtp.delay is None

    def test_delay_switch_not_above_max(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '90'
        assert smtp.delay is 60

    def test_delay_switch_invalid_range_value_no_config(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = 'fifteen, eighteen'
        assert smtp.delay is None

    def test_delay_switch_invalid_min_range_value_no_config(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = 'fifteen, 18'
        assert smtp.delay is None

    def test_delay_switch_invalid_max_range_value_no_config(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '15, eighteen'
        assert smtp.delay is None

    def test_delay_switch_invalid_range_negative_min_value(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '-10, 10'
        assert smtp.delay is None

    def test_delay_switch_invalid_range_negative_max_value(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '1, -10'
        assert smtp.delay is None

    def test_delay_switch_invalid_range_negatives(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '-10, -1'
        assert smtp.delay is None

    def test_delay_switch_range_min_higher_than_max(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '20, 10'
        assert smtp.delay is None

    def test_delay_switch_range_max_higher_than_60(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '59, 70'
        assert smtp.delay in [59, 60]

    def test_delay_switch_more_than_2(self):
        cfile = create_config(('', ))
        Config(cfile).load()
        smtp = Smtp([])
        smtp.delay = '1, 2, 3'
        assert smtp.delay is None



# Copyright (c) 2017 Kura
# Copyright (C) 2009-2015 Barry A. Warsaw
#
# This file is part of setup_helpers.py
#
# setup_helpers.py is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, version 3 of the License.
#
# setup_helpers.py is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with setup_helpers.py.  If not, see <http://www.gnu.org/licenses/>.

"""setup.py helper functions."""

import codecs
import os
import pathlib
import sys

from setuptools.command.test import test as TestCommand

__all__ = ('get_version', 'include_file', 'require_python', 'PyTest')


class PyTest(TestCommand):
    """Test command."""

    def finalize_options(self):
        """Build options."""
        TestCommand.finalize_options(self)
        self.test_args = ['--pylama', '--verbose', './blackhole', './tests']
        self.test_suite = True

    def run_tests(self):
        """Run ze tests."""
        import pytest
        sys.exit(pytest.main(self.test_args))


def require_python(minimum):
    """Python version check."""
    if sys.hexversion < minimum:
        hversion = hex(minimum)[2:]
        if len(hversion) % 2 != 0:
            hversion = '0' + hversion
        split = list(hversion)
        parts = []
        while split:
            parts.append(int(''.join((split.pop(0), split.pop(0))), 16))
        major, minor, micro, release = parts
        if release == 0xf0:
            print('Python {0}.{1}.{2} or higher is required'.format(
                major, minor, micro))
        else:
            print('Python {0}.{1}.{2} ({3}) or higher is required'.format(
                major, minor, micro, hex(release)[2:]))
        sys.exit(1)


def include_file(filename):
    """Include contents of specified file."""
    here = os.path.abspath(os.path.dirname(__file__))
    fpath = os.path.join(pathlib.PurePath(here), pathlib.PurePath(filename))
    fpath = pathlib.PurePath(fpath)
    if not os.access(fpath, os.R_OK):
        raise OSError('Cannot open {} for reading', fpath)
    return codecs.open(fpath, encoding='utf-8').read()


def get_version(filepath):
    """Return program version."""
    for line in include_file(filepath).split('\n'):
        if line.startswith('__version__'):
            _, vers = line.split('=')
            return vers.strip().replace('"', '').replace("'", '')



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import asyncio
import os
import socket
from unittest import mock

import pytest

from blackhole.streams import StreamProtocol

from ._utils import (Args, cleandir, create_config, create_file, reset)


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_client_not_connected():
    sp = StreamProtocol()
    assert sp.is_connected() is False


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_client_connected(event_loop):
    sp = StreamProtocol(loop=event_loop)
    sp.connection_made(asyncio.Transport())
    assert sp.is_connected() is True


# @pytest.mark.usefixtures('reset_conf', 'reset_daemon', 'reset_supervisor',
#         'cleandir')
# @pytest.mark.asyncio
# async def test_connection_lost_exception(event_loop):
#     sp = StreamProtocol(loop=event_loop)
#     sp.connection_made(asyncio.Transport())
#     assert sp.is_connected() is True
#     with pytest.raises(KeyError):
#         sp.connection_lost(KeyError)



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import asyncio
import unittest
from unittest import mock

import pytest

from blackhole.config import Config
from blackhole.exceptions import BlackholeRuntimeException
from blackhole.supervisor import Supervisor

from ._utils import (Args, cleandir, create_config, create_file, reset)


try:
    import ssl
except ImportError:
    ssl = None

try:
    import uvloop
    _LOOP = 'uvloop.Loop'
except ImportError:
    _LOOP = 'asyncio.unix_events._UnixSelectorEventLoop'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4():
    cfile = create_config(('listen=127.0.0.1:9999', ))
    Config(cfile).load()
    with mock.patch('socket.socket.bind'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 1


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4_tls():
    cert = create_file('cert.pem')
    key = create_file('key.key')
    cfile = create_config(('listen=127.0.0.1:9998',
                           'tls_listen=127.0.0.1:9999',
                           'tls_cert={}'.format(cert),
                           'tls_key={}'.format(key)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', True), ))
    with mock.patch('socket.socket.bind'), \
            mock.patch('ssl.create_default_context'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv6_tls():
    cert = create_file('cert.pem')
    key = create_file('key.key')
    cfile = create_config(('listen=::9998',
                           'tls_listen=:::9999',
                           'tls_cert={}'.format(cert),
                           'tls_key={}'.format(key)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', True), ))
    with mock.patch('socket.socket.bind'), \
            mock.patch('ssl.create_default_context'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4_tls_less_secure():
    cert = create_file('cert.pem')
    key = create_file('key.key')
    cfile = create_config(('listen=127.0.0.1:9998',
                           'tls_listen=127.0.0.1:9999',
                           'tls_cert={}'.format(cert),
                           'tls_key={}'.format(key)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('socket.socket.bind'), \
            mock.patch('ssl.create_default_context'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv6_tls_less_secure():
    cert = create_file('cert.pem')
    key = create_file('key.key')
    cfile = create_config(('listen=::9998',
                           'tls_listen=:::9999',
                           'tls_cert={}'.format(cert),
                           'tls_key={}'.format(key)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('socket.socket.bind'), \
            mock.patch('ssl.create_default_context'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2


key_data = """-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAtkJc/03E/Q32Oj6ixcTiY9BxWPBLaoGXqS1eSnZXIgHVfKwQ
8QyVb433hxtEYrTgrgHzFIfThf8f0ZISyPCtTXVJwBv08TeQXE2iGbHsFDKFGG9m
fDC+Cx6t8uKne81x2dCHXHeoUDKm1ae+FyUFhqMfM5Zp0AJZLueUw+8ooO5aJDCj
Jq+QgTnaFY/ZjpNK22hbRGBoUMJWUykvQMle7X52WQqrRkNIcqTvqYUCaGWqUV63
+3U9AgbuejOuZ+43MmZGdvpJSVyBVbkmeklbtD27dDrxRm4EC+LpUdpiXDIO1fLp
cj0ibuDArwZd0h6Lc7eDiOHdiv3EVlr29JZS9wIDAQABAoIBACOrVTv54mWMB/Za
L8z4vT2/m49GZ9wORLotUNjN1DuplDh0DFTAm7ZbWGguo/GVaKtlYkdQVctRpM6F
AfwlbiVMoQRh9EwJDc2iu+i/c254fKfNlYcUmvzeIjj4tsJ58KAqOXOVT3FymEgU
LhWlfzFcV0znUQoOi5msZOb7tT3ZUz8PJz8P30HaERLbLTTYR3qWdnS4ASlvfpX4
BnqPwQmi7NebNgXdlc3JhVYFtHna+iNkas7vXESw2Y1L491QpT2bgCAexs5VKe34
gnRnKzugzaX5E/3Kpyrsadp/cnIVHW7OUq+BLXvLZVJo7Y+oyMHvzpDMI87xnXQn
BqGLdsECgYEA5iCD3f7g3lTGBnfqYk1SwhTOlW9O4jsRodYtmLZHdswxdDppwaHN
oHVS1dL7xjWHMgAw7ia9sQWdi9xVLbHWAS4NiWzUlGa7Y30kKFZCspp6FdDVAl+l
nYkJe63XhPUSzIj8HGPRF0ouEwXWQG8frmZOIRco/v+MzuK1sf9ANUcCgYEAysAf
xTslUExIBGDYSAfHdCkKuPRvfspuFHbIe5+9WqEzpnSIcyQ/c/Kiy2gPRd4sCFZR
O341ljRGnGqVoKmMw11kFKtSZbQ0StoXFHojfUD2wfcX+OdeZYWIQfOA8ysDU07D
6h7pShN5sRkJJ43myLgiMvf0k2+Cb3LWbw+4jNECgYAn2r27tqaPGrNo71+iQq3w
etYEP1C0EXLOSl9/MSJTSq3N6Ufjhvgrr+X2riF4hFCzsiDHPeJgRA6y6CfAUCo1
7hneEQEH6nrNkiCu87F9aQ93s8EQNixiihcgWk0W87KbxaWT7R2LIQe1t68RXa1o
8MLbLD2nW+6w7cZ12zTthQKBgQCFCy+N3/LrTzMO1HH4SkqCmg/0MEodnD5B5XRH
UxoRJ6jZ+Kw1SjrtHfHfoggro8+nJOUnMfl9+A3BxFstUzUQWe87dPSb0nViDNyK
Tl+NaJzDyR4M8d5KpiY0iNI8PyRufUgoEGKrfj4jjTcpon8nhVTDt1vhh5HDCktW
E63ZkQKBgEr0x4bNIuCStpJwkFn6Zh8xnThCwnhH0MH0KOOzVevYTFAy8u8KbSOw
e2G8VWGmY/GoGUleHqqXLSi6UereS4TEBMh7ZzLhc5OGoljgO4dLyxvKu3eyfjGX
4KX04JK3CyPeNdL0/UIlMdTdWV17eTpfQ8hOwOVxPWJo5iiO4Knx
-----END RSA PRIVATE KEY-----"""
cert_data = """-----BEGIN CERTIFICATE-----
MIIDBjCCAe4CCQC+JlUbr0pa8zANBgkqhkiG9w0BAQsFADBFMQswCQYDVQQGEwJB
VTETMBEGA1UECAwKU29tZS1TdGF0ZTEhMB8GA1UECgwYSW50ZXJuZXQgV2lkZ2l0
cyBQdHkgTHRkMB4XDTE2MDUwMTIzMDgzN1oXDTE2MDUzMTIzMDgzN1owRTELMAkG
A1UEBhMCQVUxEzARBgNVBAgMClNvbWUtU3RhdGUxITAfBgNVBAoMGEludGVybmV0
IFdpZGdpdHMgUHR5IEx0ZDCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEB
ALZCXP9NxP0N9jo+osXE4mPQcVjwS2qBl6ktXkp2VyIB1XysEPEMlW+N94cbRGK0
4K4B8xSH04X/H9GSEsjwrU11ScAb9PE3kFxNohmx7BQyhRhvZnwwvgserfLip3vN
cdnQh1x3qFAyptWnvhclBYajHzOWadACWS7nlMPvKKDuWiQwoyavkIE52hWP2Y6T
SttoW0RgaFDCVlMpL0DJXu1+dlkKq0ZDSHKk76mFAmhlqlFet/t1PQIG7nozrmfu
NzJmRnb6SUlcgVW5JnpJW7Q9u3Q68UZuBAvi6VHaYlwyDtXy6XI9Im7gwK8GXdIe
i3O3g4jh3Yr9xFZa9vSWUvcCAwEAATANBgkqhkiG9w0BAQsFAAOCAQEALQVsOsbm
T13SX07yZ49+A0MSalX9u4QMe2L/bonYEs2/TB4dads7EPrZmhfGQ5cQlHaKp+U3
ZCGqlMMrsEWtYu3Ovc2DLJTykccb5gJV/eAvt8siBH1+8EVuwptz4KBAPbosxtaZ
ITeGcU/K5Sg6VcVqmEB1Bam7jYyk91f42mtqINS3QmtBhGDCDyGLu13N0Tcli4Ex
tG2AOVmjWihvNmd21SLOJoQDn+8C04j5sTWVfLxRbfwFyP/frhAYN4EfHVHLjHf/
n3ZYIxa8EMbZyJNQfipZeBUrAyYtjgHCRfgM2t5TscMfB1ewrQ/iDSNE5a65vYOC
VMr7S68leXh3Aw==
-----END CERTIFICATE-----"""
dhparams_data = """-----BEGIN DH PARAMETERS-----
MEYCQQDkC0FDza65yPJj8HeXOpoR315XGC+/5uaJpw5CevVwAmRns3TEuFwTal5H
MWsNrHaep9EbyJ00JW/cSoaECAYLAgEC
-----END DH PARAMETERS-----"""


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4_tls_dhparams():
    cert = create_file('test.pem', cert_data)
    key = create_file('test.key', key_data)
    dhparams = create_file('dhparams.pem', dhparams_data)
    cfile = create_config(('listen=127.0.0.1:9998',
                           'tls_listen=127.0.0.1:9999',
                           'tls_cert={}'.format(cert),
                           'tls_key={}'.format(key),
                           'tls_dhparams={}'.format(dhparams)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('socket.socket.bind'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2


@unittest.skipIf(ssl is None, 'No ssl module')
@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv6_tls_dhparams():
    cert = create_file('test.pem', cert_data)
    key = create_file('test.key', key_data)
    dhparams = create_file('dhparams.pem', dhparams_data)
    cfile = create_config(('listen=:::9998',
                           'tls_listen=:::9999',
                           'tls_cert={}'.format(cert),
                           'tls_key={}'.format(key),
                           'tls_dhparams={}'.format(dhparams)))
    conf = Config(cfile).load()
    conf.args = Args((('less_secure', False), ))
    with mock.patch('socket.socket.bind'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2
    assert supervisor.socks[1]['ssl'] is not None


@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv6():
    cfile = create_config(('listen=:::9999', ))
    Config(cfile).load()
    with mock.patch('socket.socket.bind'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 1


@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4_and_ipv6():
    cfile = create_config(('listen=:9999, :::9999', ))
    Config(cfile).load()
    supervisor = Supervisor()
    with mock.patch('socket.socket.bind'):
        supervisor = Supervisor()
    assert len(supervisor.socks) == 2


@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4_fail():
    cfile = create_config(('listen=:9999', ))
    Config(cfile).load()
    with mock.patch('socket.socket.bind', side_effect=OSError), \
            pytest.raises(BlackholeRuntimeException):
        Supervisor()


@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv6_fail():
    cfile = create_config(('listen=:::9999', ))
    Config(cfile).load()
    with mock.patch('socket.socket.bind', side_effect=OSError), \
            pytest.raises(BlackholeRuntimeException):
        Supervisor()


@pytest.mark.usefixtures('reset', 'cleandir')
def test_spawn_ipv4_and_ipv6_fail():
    cfile = create_config(('listen=:9999, :::9999', ))
    Config(cfile).load()
    with mock.patch('socket.socket.bind', side_effect=OSError), \
            pytest.raises(BlackholeRuntimeException):
        Supervisor()


@pytest.mark.usefixtures('reset', 'cleandir')
def test_create():
    cfile = create_config(('listen=:9999, :::9999', 'workers=2', ))
    Config(cfile).load()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    with mock.patch('socket.socket.bind'), \
            mock.patch('blackhole.worker.Worker.start'):
        supervisor = Supervisor(loop=loop)
        supervisor.start_workers()
        assert len(supervisor.workers) == 2
    loop.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run():
    cfile = create_config(('listen=:9999, :::9999', 'workers=2', ))
    Config(cfile).load()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    with mock.patch('socket.socket.bind'), \
            mock.patch('blackhole.worker.Worker.start'):
        supervisor = Supervisor(loop)
        with mock.patch('{0}.run_forever'.format(_LOOP)):
            supervisor.run()
        assert len(supervisor.workers) == 2
    supervisor.loop.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
def test_stop():
    cfile = create_config(('listen=:9999, :::9999', 'workers=2', ))
    Config(cfile).load()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    with mock.patch('socket.socket.bind'), \
            mock.patch('blackhole.worker.Worker.start'):
        supervisor = Supervisor(loop)
        supervisor.start_workers()
        assert len(supervisor.workers) == 2
        with mock.patch('blackhole.worker.Worker.stop') as mock_stop, \
                pytest.raises(SystemExit) as err:
            supervisor.stop()
    assert mock_stop.call_count == 2
    assert str(err.value) == '0'
    supervisor.loop.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
def test_stop_runtime_error():
    cfile = create_config(('listen=:9999, :::9999', 'workers=2', ))
    Config(cfile).load()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    with mock.patch('socket.socket.bind'), \
            mock.patch('blackhole.worker.Worker.start'):
        supervisor = Supervisor(loop)
        supervisor.start_workers()
        assert len(supervisor.workers) == 2
        with mock.patch('blackhole.worker.Worker.stop') as mock_stop, \
                mock.patch('{0}.stop'.format(_LOOP),
                           side_effect=RuntimeError) as mock_rt, \
                pytest.raises(SystemExit) as err:
            supervisor.stop()
    assert mock_rt.call_count == 1
    assert mock_stop.call_count == 2
    assert str(err.value) == '0'
    supervisor.loop.stop()



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

from io import StringIO
from unittest import mock

import pytest

from blackhole.utils import get_version, mailname, message_id

from ._utils import (Args, cleandir, create_config, create_file, reset)


@pytest.mark.usefixtures('reset', 'cleandir')
def test_mail_name_file():
    check_value = 'file.blackhole.io'
    with mock.patch('os.access', return_value=True), \
            mock.patch('codecs.open', return_value=StringIO(check_value)):
        mn = mailname()
        assert mn == check_value


@pytest.mark.usefixtures('reset', 'cleandir')
def test_mail_name_socket():
    check_value = 'socket.blackhole.io'
    with mock.patch('os.access', return_value=False), \
            mock.patch('socket.getfqdn', return_value='socket.blackhole.io'):
        mn = mailname()
    assert mn == check_value


@pytest.mark.usefixtures('reset', 'cleandir')
def test_mail_name_file_length_0():
    mnfile = create_file('mailname')
    check_value = 'socket.blackhole.io'
    with mock.patch('os.access', return_value=True), \
            mock.patch('socket.getfqdn', return_value='socket.blackhole.io'):
        mn = mailname(mnfile)
    assert mn == check_value


@pytest.mark.usefixtures('reset', 'cleandir')
def test_mail_name_file_garbage():
    mnfile = create_file('mailname', '            \n     ')
    check_value = 'socket.blackhole.io'
    with mock.patch('os.access', return_value=True), \
            mock.patch('socket.getfqdn', return_value='socket.blackhole.io'):
        mn = mailname(mnfile)
    assert mn == check_value


@pytest.mark.usefixtures('reset', 'cleandir')
def test_message_id():
    with mock.patch('time.monotonic',
                    return_value=1463290829.4173775) as mock_time, \
        mock.patch('os.getpid', return_value=9000) as mock_getpid, \
            mock.patch('random.getrandbits',
                       return_value=17264867586200823825) as mock_randbits:
        ex_mid = '<{}.{}.{}@blackhole.io>'.format(int(1463290829.4173775 *
                                                  100),
                                                  9000, 17264867586200823825)
        assert message_id('blackhole.io') == ex_mid
    assert mock_time.called is True
    assert mock_getpid.called is True
    assert mock_randbits.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_version():
    version_file = create_file('version.py', '__version__ = "9.9.9"')
    with mock.patch('os.path.join', return_value=version_file):
        assert get_version() == '9.9.9'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_version_no_access():
    with mock.patch('os.access', return_value=False), \
            pytest.raises(OSError) as err:
        get_version()
    assert str(err.value) == 'Cannot open __init__.py file for reading'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_version_invalid_version_split():
    version_file = create_file('version.py', '__version__')
    with mock.patch('os.path.join', return_value=version_file), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'Cannot extract version from __version__'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_version_invalid_version():
    version_file_a = create_file('versiona.py', '__version__ = a.1')
    with mock.patch('os.path.join', return_value=version_file_a), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'a.1 is not a valid version number'
    version_file_b = create_file('versionb.py', '__version__ = a.1.1')
    with mock.patch('os.path.join', return_value=version_file_b), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'a.1.1 is not a valid version number'
    version_file_c = create_file('versionc.py', '__version__ = 1.a.1')
    with mock.patch('os.path.join', return_value=version_file_c), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == '1.a.1 is not a valid version number'
    version_file_d = create_file('versiond.py', '__version__ = 1.1.a')
    with mock.patch('os.path.join', return_value=version_file_d), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == '1.1.a is not a valid version number'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_version_version_not_found():
    version_file = create_file('version.py', 'version = "abc"')
    with mock.patch('os.path.join', return_value=version_file), \
            pytest.raises(AssertionError) as err:
        get_version()
    assert str(err.value) == 'No __version__ assignment found'



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import asyncio
import os
import socket
from unittest import mock

import pytest

from blackhole import protocols
from blackhole.child import Child
from blackhole.control import _socket
from blackhole.streams import StreamProtocol

from ._utils import (Args, cleandir, create_config, create_file, reset)


try:
    import uvloop
    _LOOP = 'uvloop.Loop'
except ImportError:
    _LOOP = 'asyncio.unix_events._UnixSelectorEventLoop'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_initiation():
    Child('', '', [], '1')


@pytest.mark.usefixtures('reset', 'cleandir')
def test_start():
    socks = [{'sock': None, 'ssl': None}, {'sock': None, 'ssl': 'abc'}]
    child = Child('', '', socks, '1')
    with mock.patch('asyncio.Task'), \
        mock.patch('blackhole.child.Child.heartbeat'), \
        mock.patch('{0}.run_forever'.format(_LOOP)), \
        mock.patch('blackhole.child.Child.stop'), \
            mock.patch('os._exit') as mock_exit:
        child.start()
    assert mock_exit.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
def test_stop():
    socks = [{'sock': None, 'ssl': None}, {'sock': None, 'ssl': 'abc'}]
    child = Child('', '', socks, '1')
    child.loop = mock.MagicMock()
    child.clients.append(mock.MagicMock())
    child.servers.append(mock.MagicMock())
    child.heartbeat_task = mock.MagicMock()
    child.server_task = mock.MagicMock()

    with mock.patch('os._exit') as mock_exit, \
            mock.patch('{0}.run_until_complete'.format(_LOOP)):
        child.stop()
    assert mock_exit.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
def test_stop_runtime_exception():
    socks = [{'sock': None, 'ssl': None}, {'sock': None, 'ssl': 'abc'}]
    child = Child('', '', socks, '1')
    child.loop = mock.MagicMock()
    child.clients.append(mock.MagicMock())
    child.servers.append(mock.MagicMock())
    child.heartbeat_task = mock.MagicMock()
    child.server_task = mock.MagicMock()

    with mock.patch('os._exit') as mock_exit, \
            mock.patch('{0}.stop'.format(_LOOP), side_effect=RuntimeError):
        child.stop()
    assert mock_exit.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_start_child_loop(event_loop):
    sock = _socket('127.0.0.1', 0, socket.AF_INET)
    socks = ({'sock': sock, 'ssl': None}, )
    child = Child('', '', socks, '1')
    child.loop = event_loop
    await child._start()
    assert len(child.servers) == 1
    for server in child.servers:
        server.close()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_child_heartbeat_not_started(event_loop):
    up_read, up_write = os.pipe()
    down_read, down_write = os.pipe()
    os.close(up_write)
    os.close(down_read)
    child = Child(up_read, down_write, [], '1')
    child.loop = event_loop
    child._started = False
    with mock.patch('asyncio.Task') as mock_task, \
        mock.patch('blackhole.child.Child._start') as mock_start, \
            mock.patch('blackhole.child.Child.stop') as mock_stop:
        await child.heartbeat()
    assert mock_task.called is True
    assert mock_start.called is True
    assert mock_stop.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_child_heartbeat_started(event_loop):
    up_read, up_write = os.pipe()
    down_read, down_write = os.pipe()
    os.close(up_write)
    os.close(down_read)
    child = Child(up_read, down_write, [], '1')
    child.loop = event_loop
    child._started = True
    sp = StreamProtocol()
    sp.reader = asyncio.StreamReader()

    async def reset():
        sp.reader.feed_data(protocols.PING)
        child._started = False
    reset_task = asyncio.Task(reset())
    with mock.patch('blackhole.streams.StreamProtocol', return_value=sp), \
        mock.patch('asyncio.Task') as mock_task, \
        mock.patch('blackhole.child.Child._start') as mock_start, \
            mock.patch('blackhole.child.Child.stop') as mock_stop:
        await child.heartbeat()
    reset_task.cancel()
    assert mock_task.called is True
    assert mock_start.called is True
    assert mock_stop.called is True



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import asyncio
import inspect
from smtplib import SMTP, SMTPNotSupportedError, SMTPServerDisconnected
import socket
import threading
import time
import unittest
from unittest import mock

import pytest

from blackhole.config import Config
from blackhole.control import _socket
from blackhole.smtp import Smtp

from ._utils import (Args, cleandir, create_config, create_file, reset)


try:
    import asyncio
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


@pytest.mark.usefixtures('reset', 'cleandir')
def test_initiation():
    cfile = create_config(('', ))
    with mock.patch('os.access', return_value=False), \
            mock.patch('socket.getfqdn', return_value='a.blackhole.io'):
        conf = Config(cfile)
    conf.load()
    smtp = Smtp([])
    assert smtp.fqdn == 'a.blackhole.io'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_auth_mechanisms():
    smtp = Smtp([])
    assert smtp.get_auth_members() == ['CRAM-MD5', 'LOGIN', 'PLAIN']


@pytest.mark.usefixtures('reset', 'cleandir')
def test_handler_lookup():
    smtp = Smtp([])
    assert smtp.lookup_handler('AUTH CRAM-MD5') == smtp.auth_CRAM_MD5
    assert smtp.lookup_handler('AUTH LOGIN') == smtp.auth_LOGIN
    assert smtp.lookup_handler('AUTH PLAIN') == smtp.auth_PLAIN
    assert smtp.lookup_handler('AUTH') == smtp.auth_UNKNOWN
    assert smtp.lookup_handler('AUTH KURA') == smtp.auth_UNKNOWN
    assert smtp.lookup_handler('HELP') == smtp.do_HELP
    assert smtp.lookup_handler('DATA') == smtp.do_DATA
    assert smtp.lookup_handler('EHLO') == smtp.do_EHLO
    assert smtp.lookup_handler('ETRN') == smtp.do_ETRN
    assert smtp.lookup_handler('EXPN') == smtp.do_EXPN
    assert smtp.lookup_handler('HELO') == smtp.do_HELO
    assert smtp.lookup_handler('MAIL') == smtp.do_MAIL
    assert smtp.lookup_handler('NOOP') == smtp.do_NOOP
    assert smtp.lookup_handler('QUIT') == smtp.do_QUIT
    assert smtp.lookup_handler('RCPT') == smtp.do_RCPT
    assert smtp.lookup_handler('RSET') == smtp.do_RSET
    assert smtp.lookup_handler('VRFY') == smtp.do_VRFY
    assert smtp.lookup_handler('STARTTLS') == smtp.do_STARTTLS
    assert smtp.lookup_handler('KURA') == smtp.do_UNKNOWN
    assert smtp.lookup_handler('') == smtp.do_UNKNOWN
    assert smtp.lookup_handler('STARTTLS') == smtp.do_STARTTLS
    assert smtp.lookup_handler('HELP DATA') == smtp.help_DATA
    assert smtp.lookup_handler('HELP EHLO') == smtp.help_EHLO
    assert smtp.lookup_handler('HELP ETRN') == smtp.help_ETRN
    assert smtp.lookup_handler('HELP EXPN') == smtp.help_EXPN
    assert smtp.lookup_handler('HELP HELO') == smtp.help_HELO
    assert smtp.lookup_handler('HELP MAIL') == smtp.help_MAIL
    assert smtp.lookup_handler('HELP NOOP') == smtp.help_NOOP
    assert smtp.lookup_handler('HELP QUIT') == smtp.help_QUIT
    assert smtp.lookup_handler('HELP RCPT') == smtp.help_RCPT
    assert smtp.lookup_handler('HELP RSET') == smtp.help_RSET
    assert smtp.lookup_handler('HELP VRFY') == smtp.help_VRFY
    assert smtp.lookup_handler('HELP KURA') == smtp.help_UNKNOWN


@pytest.mark.usefixtures('reset', 'cleandir')
def test_unknown_handlers():
    # Protection against adding/removing without updating tests
    verbs = ['do_DATA', 'do_EHLO', 'do_ETRN', 'do_EXPN', 'do_HELO', 'do_HELP',
             'do_MAIL', 'do_NOOP', 'do_NOT_IMPLEMENTED', 'do_QUIT', 'do_RCPT',
             'do_RSET', 'do_STARTTLS', 'do_UNKNOWN', 'do_VRFY']
    helps = ['help_AUTH', 'help_DATA', 'help_EHLO', 'help_ETRN', 'help_EXPN',
             'help_HELO', 'help_MAIL', 'help_NOOP', 'help_QUIT', 'help_RCPT',
             'help_RSET', 'help_UNKNOWN', 'help_VRFY']
    auths = ['auth_CRAM_MD5', 'auth_LOGIN', 'auth_PLAIN', 'auth_UNKNOWN']
    smtp = Smtp([])
    for mem in inspect.getmembers(smtp, inspect.ismethod):
        f, _ = mem
        if f.startswith('do_'):
            assert f in verbs
        if f.startswith('help_'):
            assert f in helps
        if f.startswith('auth_'):
            assert f in auths


@pytest.mark.usefixtures('reset', 'cleandir')
class Controller:

    def __init__(self, sock=None):
        if sock is not None:
            self.sock = sock
        else:
            self.sock = _socket('127.0.0.1', 0, socket.AF_INET)
        self.loop = asyncio.new_event_loop()
        self.server = None
        self._thread = None

    def _run(self, ready_event):
        asyncio.set_event_loop(self.loop)
        conf = Config(None)
        conf.mailname = 'blackhole.io'
        _server = self.loop.create_server(lambda: Smtp([]),
                                          sock=self.sock)
        self.server = self.loop.run_until_complete(_server)
        self.loop.call_soon(ready_event.set)
        self.loop.run_forever()
        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())
        self.loop.close()
        self.server = None

    def start(self):
        ready_event = threading.Event()
        self._thread = threading.Thread(target=self._run, args=(ready_event, ))
        self._thread.daemon = True
        self._thread.start()
        ready_event.wait()

    def stop(self):
        assert self._thread is not None, 'SMTP daemon not running'
        self.loop.call_soon_threadsafe(self._stop)
        self._thread.join()
        self._thread = None

    def _stop(self):
        self.loop.stop()
        for task in asyncio.Task.all_tasks(self.loop):
            task.cancel()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_mode_directive(event_loop, unused_tcp_port):
    cfile = create_config(('listen=:{} mode=bounce'.format(unused_tcp_port), ))
    conf = Config(cfile).load()
    sock = _socket('127.0.0.1', unused_tcp_port, socket.AF_INET)
    controller = Controller(sock)
    controller.start()
    conf.flags_from_listener('127.0.0.1', unused_tcp_port)
    host, port = sock.getsockname()
    with SMTP(host, port) as client:
        msg = ['From: kura@example.com', 'To: kura@example.com',
               'Subject: Test', 'X-Blackhole-Mode: accept',
               'X-Blackhole-Delay: 5', '', 'Testing 1, 2, 3']
        msg = '\n'.join(msg)
        start = time.time()
        code, resp = client.data(msg.encode('utf-8'))
        stop = time.time()
        assert code in [450, 451, 452, 458, 521, 550, 551, 552, 553, 571]
        assert round(stop - start) in (0, 1)
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_delay_directive(event_loop, unused_tcp_port):
    cfile = create_config(('listen=:{} delay=5'.format(unused_tcp_port), ))
    conf = Config(cfile).load()
    sock = _socket('127.0.0.1', unused_tcp_port, socket.AF_INET)
    controller = Controller(sock)
    controller.start()
    conf.flags_from_listener('127.0.0.1', unused_tcp_port)
    host, port = sock.getsockname()
    with SMTP(host, port) as client:
        msg = ['From: kura@example.com', 'To: kura@example.com',
               'Subject: Test', 'X-Blackhole-Mode: bounce',
               'X-Blackhole-Delay: 30', '', 'Testing 1, 2, 3']
        msg = '\n'.join(msg)
        start = time.time()
        code, resp = client.data(msg.encode('utf-8'))
        stop = time.time()
        assert code == 250
        assert round(stop - start) in (4, 5, 6)
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_mode_and_delay_directive(event_loop, unused_tcp_port):
    cfile = create_config(('listen=:{} delay=5 mode=bounce'.format(unused_tcp_port), ))
    conf = Config(cfile).load()
    sock = _socket('127.0.0.1', unused_tcp_port, socket.AF_INET)
    controller = Controller(sock)
    controller.start()
    conf.flags_from_listener('127.0.0.1', unused_tcp_port)
    host, port = sock.getsockname()
    with SMTP(host, port) as client:
        msg = ['From: kura@example.com', 'To: kura@example.com',
               'Subject: Test', 'X-Blackhole-Mode: accept',
               'X-Blackhole-Delay: 30', '', 'Testing 1, 2, 3']
        msg = '\n'.join(msg)
        start = time.time()
        code, resp = client.data(msg.encode('utf-8'))
        stop = time.time()
        assert code in [450, 451, 452, 458, 521, 550, 551, 552, 553, 571]
        assert round(stop - start) in (4, 5, 6)
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_timeout(event_loop):
    cfile = create_config(('timeout=5', ))
    Config(cfile).load()
    controller = Controller()
    controller.start()
    host, port = controller.sock.getsockname()
    with SMTP(host, port) as client:
        await asyncio.sleep(8)
        code, resp = client.helo('example.com')
        assert code == 421
        assert resp == b'Timeout'
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_delay(event_loop):
    cfile = create_config(('timeout=10', ))
    Config(cfile).load()
    controller = Controller()
    controller.start()
    host, port = controller.sock.getsockname()
    with SMTP(host, port) as client:
        msg = ['From: kura@example.com', 'To: kura@example.com',
               'Subject: Test', 'X-Blackhole-Mode: accept',
               'X-Blackhole-Delay: 5', '', 'Testing 1, 2, 3']
        msg = '\n'.join(msg)
        start = time.time()
        code, resp = client.data(msg.encode('utf-8'))
        stop = time.time()
        assert code == 250
        assert resp.startswith(b'2.0.0 OK: queued as')
        assert round(stop - start) in (4, 5, 6)
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_delayed_bounce(event_loop):
    cfile = create_config(('timeout=10', ))
    Config(cfile).load()
    controller = Controller()
    controller.start()
    host, port = controller.sock.getsockname()
    with SMTP(host, port) as client:
        msg = ['From: kura@example.com', 'To: kura@example.com',
               'Subject: Test', 'X-Blackhole-Mode: bounce',
               'X-Blackhole-Delay: 5', '', 'Testing 1, 2, 3']
        msg = '\n'.join(msg)
        start = time.time()
        code, resp = client.data(msg.encode('utf-8'))
        stop = time.time()
        assert code in [450, 451, 452, 458, 521, 550, 551, 552, 553, 571]
        assert round(stop - start) in (4, 5, 6)
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_delay_range(event_loop):
    cfile = create_config(('timeout=10', ))
    Config(cfile).load()
    controller = Controller()
    controller.start()
    host, port = controller.sock.getsockname()
    with SMTP(host, port) as client:
        msg = ['From: kura@example.com', 'To: kura@example.com',
               'Subject: Test', 'X-Blackhole-Mode: accept',
               'X-Blackhole-Delay: 2, 4', '', 'Testing 1, 2, 3']
        msg = '\n'.join(msg)
        start = time.time()
        code, resp = client.data(msg.encode('utf-8'))
        stop = time.time()
        assert code == 250
        assert resp.startswith(b'2.0.0 OK: queued as')
        assert round(stop - start) in (2, 3, 4)
    controller.stop()


@pytest.mark.usefixtures('reset', 'cleandir')
class TestSmtp(unittest.TestCase):

    def setUp(self):
        cfile = create_config(('timeout=5', 'max_message_size=102400'))
        Config(cfile).load()
        controller = Controller()
        controller.start()
        self.host, self.port = controller.sock.getsockname()
        self.addCleanup(controller.stop)

    def test_helo(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.helo('example.com')
            assert code == 250
            assert resp == b'OK'

    def test_ehlo(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.ehlo('example.com')
            assert code == 250
            # resps = resp.decode('utf-8').split('\n')
            eresp = ('blackhole.io', 'HELP', 'PIPELINING',
                     'AUTH CRAM-MD5 LOGIN PLAIN', 'SIZE 102400', 'VRFY',
                     'ETRN', 'ENHANCEDSTATUSCODES', '8BITMIME', 'SMTPUTF8',
                     'EXPN', 'DSN')
            assert resp == '\n'.join(eresp).encode('utf-8')

    def test_mail(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.mail('kura@example.com')
            assert code == 250
            assert resp == b'2.1.0 OK'

    def test_mail_size_ok(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.mail('kura@example.com SIZE=1')
            assert code == 250
            assert resp == b'2.1.0 OK'

    def test_mail_size_ok_and_mime(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.mail('kura@example.com SMTPUTF8 SIZE=1024')
            assert code == 250
            assert resp == b'2.1.0 OK'
        with SMTP(self.host, self.port) as client:
            code, resp = client.mail('kura@example.com BODY=7BIT SIZE=1024')
            assert code == 250
            assert resp == b'2.1.0 OK'
        with SMTP(self.host, self.port) as client:
            code, resp = client.mail('kura@example.com BODY=8BITMIME '
                                     'SIZE=1024')
            assert code == 250
            assert resp == b'2.1.0 OK'

    def test_mail_size_too_large(self):
        with SMTP(self.host, self.port) as client:
            msg = 'MAIL FROM: kura@example.com SIZE=10240000000'
            code, resp = client.docmd(msg)
            assert code == 552
            assert resp == b'Message size exceeds fixed maximum message size'

    def test_mail_size_too_large_and_mime(self):
        with SMTP(self.host, self.port) as client:
            msg = 'MAIL FROM: kura@example.com SMTPUTF8 SIZE=10240000000'
            code, resp = client.docmd(msg)
            assert code == 552
            assert resp == b'Message size exceeds fixed maximum message size'
        with SMTP(self.host, self.port) as client:
            msg = 'MAIL FROM: kura@example.com BODY=7BIT SIZE=10240000000'
            code, resp = client.docmd(msg)
            assert code == 552
            assert resp == b'Message size exceeds fixed maximum message size'
        with SMTP(self.host, self.port) as client:
            msg = 'MAIL FROM: kura@example.com BODY=8BITMIME SIZE=10240000000'
            code, resp = client.docmd(msg)
            assert code == 552
            assert resp == b'Message size exceeds fixed maximum message size'

    def test_mail_smtputf8(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('MAIL FROM: kura@example.com SMTPUTF8')
            assert code == 250
            assert resp == b'2.1.0 OK'

    def test_mail_7bit(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('MAIL FROM: kura@example.com BODY=7BIT')
            assert code == 250
            assert resp == b'2.1.0 OK'

    def test_mail_8bitmime(self):
        with SMTP(self.host, self.port) as client:
            msg = 'MAIL FROM: kura@example.com BODY=8BITMIME'
            code, resp = client.docmd(msg)
            assert code == 250
            assert resp == b'2.1.0 OK'

    def test_rcpt(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.rcpt('kura@example.com')
            assert code == 250
            assert resp == b'2.1.5 OK'

    def test_data(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.data(b'testing 1, 2, 3')
            assert code == 250
            assert resp.startswith(b'2.0.0 OK: queued as')

    def test_data_fail(self):
        with SMTP(self.host, self.port) as client:
            msg = ['From: kura@example.com', 'To: kura@example.com',
                   'Subject: Test', 'X-Blackhole-Mode: bounce', '',
                   'Testing 1, 2, 3']
            msg = '\n'.join(msg)
            code, resp = client.data(msg.encode('utf-8'))
            assert code in [450, 451, 452, 458, 521, 550, 551, 552, 553, 571]

    def test_data_random(self):
        with SMTP(self.host, self.port) as client:
            msg = ['From: kura@example.com', 'To: kura@example.com',
                   'Subject: Test', 'X-Blackhole-Mode: random', '',
                   'Testing 1, 2, 3']
            msg = '\n'.join(msg)
            code, resp = client.data(msg.encode('utf-8'))
            assert code in [250, 450, 451, 452, 458, 521, 550, 551, 552, 553,
                            571]

    def test_data_accept(self):
        with SMTP(self.host, self.port) as client:
            msg = ['From: kura@example.com', 'To: kura@example.com',
                   'Subject: Test', 'X-Blackhole-Mode: accept', '',
                   'Testing 1, 2, 3']
            msg = '\n'.join(msg)
            code, resp = client.data(msg.encode('utf-8'))
            assert code == 250
            assert resp.startswith(b'2.0.0 OK: queued as')

    def test_rset(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.rset()
            assert code == 250
            assert resp == b'2.0.0 OK'

    def test_noop(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.noop()
            assert code == 250
            assert resp == b'2.0.0 OK'

    def test_vrfy(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.verify('kura@example.com')
            assert code == 252
            assert resp == b'2.0.0 Will attempt delivery'

    def test_vrfy_pass(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.verify('pass=kura@example.com')
            assert code == 250
            assert resp == b'2.0.0 <pass=kura@example.com> OK'

    def test_vrfy_fail(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.verify('fail=kura@example.com')
            assert code == 550
            assert resp == b'5.7.1 <fail=kura@example.com> unknown'

    def test_expn_no_list(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.expn('')
            assert code == 550
            assert resp == b'Not authorised'

    def test_expn_fail(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.expn('fail=kura@example.com')
            assert code == 550
            assert resp == b'Not authorised'

    def test_expn_list1(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.expn('list1')
            eresp = ('Shadow <shadow@blackhole.io>',
                     'Wednesday <wednesday@blackhole.io>',
                     'Low-key Liesmith <low-key.liesmith@blackhole.io>')
            assert code == 250
            assert resp == '\n'.join(eresp).encode('utf-8')

    def test_expn_list2(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.expn('list2')
            eresp = ('Jim Holden <jim.holden@blackhole.io>',
                     'Naomi Nagata <naomi.nagata@blackhole.io>',
                     'Alex Kamal <alex.kamal@blackhole.io>',
                     'Amos Burton <amos.burton@blackhole.io>')
            assert code == 250
            assert resp == '\n'.join(eresp).encode('utf-8')

    def test_expn_list3(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.expn('list3')
            eresp = ('Takeshi Kovacs <takeshi.kovacs@blackhole.io>',
                     'Laurens Bancroft <laurens.bancroft@blackhole.io>',
                     'Kristin Ortega <kristin.ortega@blackhole.io>',
                     'Quellcrist Falconer <quellcrist.falconer@blackhole.io>',
                     'Virginia Vidaura <virginia.vidaura@blackhole.io>',
                     'Reileen Kawahara <reileen.kawahara@blackhole.io>')
            assert code == 250
            assert resp == '\n'.join(eresp).encode('utf-8')

    def test_expn_all(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.expn('all')
            eresp = ('Takeshi Kovacs <takeshi.kovacs@blackhole.io>',
                     'Laurens Bancroft <laurens.bancroft@blackhole.io>',
                     'Kristin Ortega <kristin.ortega@blackhole.io>',
                     'Quellcrist Falconer <quellcrist.falconer@blackhole.io>',
                     'Virginia Vidaura <virginia.vidaura@blackhole.io>',
                     'Reileen Kawahara <reileen.kawahara@blackhole.io>',
                     'Jim Holden <jim.holden@blackhole.io>',
                     'Naomi Nagata <naomi.nagata@blackhole.io>',
                     'Alex Kamal <alex.kamal@blackhole.io>',
                     'Amos Burton <amos.burton@blackhole.io>',
                     'Shadow <shadow@blackhole.io>',
                     'Wednesday <wednesday@blackhole.io>',
                     'Low-key Liesmith <low-key.liesmith@blackhole.io>')
            assert code == 250
            resps = resp.decode('utf-8').split('\n')
            assert sorted(resps) == sorted(eresp)

    def test_etrn(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('ETRN')
            assert code == 250
            assert resp == b'Queueing started'

    def test_quit(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.quit()
            assert code == 221
            assert resp == b'2.0.0 Goodbye'

    def test_starttls(self):
        with SMTP(self.host, self.port) as client:
            with pytest.raises(SMTPNotSupportedError):
                code, resp = client.starttls()
            code, resp = client.docmd('STARTTLS')
            assert code == 500
            assert resp == b'Not implemented'

    def test_unknown(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('KURA')
            assert code == 502
            assert resp == b'5.5.2 Command not recognised'

    def test_help(self):
        with SMTP(self.host, self.port) as client:
            eresp = ('Supported commands: AUTH DATA EHLO ETRN EXPN HELO MAIL '
                     'NOOP QUIT RCPT RSET VRFY')
            assert client.help() == eresp.encode('utf-8')

    def test_help_auth(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('AUTH')
            assert resp == b'Syntax: AUTH CRAM-MD5 LOGIN PLAIN'

    def test_help_data(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('DATA')
            assert resp == b'Syntax: DATA'

    def test_help_ehlo(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('EHLO')
            assert resp == b'Syntax: EHLO domain.tld'

    def test_help_etrn(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('ETRN')
            assert resp == b'Syntax: ETRN'

    def test_help_expn(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('EXPN')
            assert resp == b'Syntax: EXPN <list1 | list2 | list3 | all>'

    def test_help_helo(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('HELO')
            assert resp == b'Syntax: HELO domain.tld'

    def test_help_mail(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('MAIL')
            assert resp == b'Syntax: MAIL FROM: <address>'

    def test_help_noop(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('NOOP')
            assert resp == b'Syntax: NOOP'

    def test_help_quit(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('QUIT')
            assert resp == b'Syntax: QUIT'

    def test_help_rcpt(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('RCPT')
            assert resp == b'Syntax: RCPT TO: <address>'

    def test_help_rset(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('RSET')
            assert resp == b'Syntax: RSET'

    def test_help_unknown(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('HELP', 'KURA')
            eresp = ('Supported commands: AUTH DATA EHLO ETRN EXPN HELO MAIL '
                     'NOOP QUIT RCPT RSET VRFY')
            assert code == 501
            assert resp == eresp.encode('utf-8')

    def test_help_vrfy(self):
        with SMTP(self.host, self.port) as client:
            resp = client.help('VRFY')
            assert resp == b'Syntax: VRFY <address>'

    def test_auth_login(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'LOGIN')
            assert code == 334
            assert resp == b'VXNlcm5hbWU6'
            code, resp = client.docmd('test')
            assert code == 235
            assert resp == b'2.7.0 Authentication successful'

    def test_auth_login_fail(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'LOGIN')
            assert code == 334
            assert resp == b'VXNlcm5hbWU6'
            code, resp = client.docmd('fail=test')
            assert code == 535
            assert resp == b'5.7.8 Authentication failed'

    def test_auth_login_pass(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'LOGIN')
            assert code == 334
            assert resp == b'VXNlcm5hbWU6'
            code, resp = client.docmd('pass=test')
            assert code == 235
            assert resp == b'2.7.0 Authentication successful'

    def test_auth_cram_md5(self):
        with SMTP(self.host, self.port) as client:
            client.user = 'test'
            client.password = 'test'
            code, resp = client.auth('CRAM-MD5', client.auth_cram_md5)
            assert code == 235
            assert resp == b'2.7.0 Authentication successful'

    def test_auth_cram_md5_fail(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'CRAM-MD5')
            assert code == 334
            code, resp = client.docmd('fail=test')
            assert code == 535
            assert resp == b'5.7.8 Authentication failed'

    def test_auth_cram_md5_pass(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'CRAM-MD5')
            assert code == 334
            code, resp = client.docmd('pass=test')
            assert code == 235
            assert resp == b'2.7.0 Authentication successful'

    def test_auth_plain(self):
        with SMTP(self.host, self.port) as client:
            client.user = 'test'
            client.password = 'test'
            code, resp = client.auth('PLAIN', client.auth_plain)
            assert code == 235
            assert resp == b'2.7.0 Authentication successful'

    def test_auth_plain_fail(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'PLAIN')
            assert code == 334
            code, resp = client.docmd('fail=test')
            assert code == 535
            assert resp == b'5.7.8 Authentication failed'

    def test_auth_plain_pass(self):
        with SMTP(self.host, self.port) as client:
            with SMTP(self.host, self.port) as client:
                code, resp = client.docmd('AUTH', 'PLAIN')
                assert code == 334
                code, resp = client.docmd('pass=test')
                assert code == 235
                assert resp == b'2.7.0 Authentication successful'

    def test_auth_plain_fail_oneline(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'PLAIN fail=test')
            assert code == 535
            assert resp == b'5.7.8 Authentication failed'

    def test_auth_plain_pass_oneline(self):
        with SMTP(self.host, self.port) as client:
            with SMTP(self.host, self.port) as client:
                code, resp = client.docmd('AUTH', 'PLAIN pass=test')
                assert code == 235
                assert resp == b'2.7.0 Authentication successful'

    def test_auth_unknown(self):
        with SMTP(self.host, self.port) as client:
            code, resp = client.docmd('AUTH', 'KURA')
            assert code == 501
            assert resp == b'5.5.4 Syntax: AUTH mechanism'

    def test_too_many_unknown_commands(self):
        with SMTP(self.host, self.port) as client, \
                pytest.raises(SMTPServerDisconnected):
            for _ in range(11):
                code, resp = client.docmd('KURA')
            assert code == 502
            assert resp == b'5.5.3 Too many unknown commands'



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import logging
import os
from unittest import mock

import pytest

from blackhole.application import blackhole_config, run
from blackhole.config import Config
from blackhole.exceptions import (BlackholeRuntimeException, ConfigException,
                                  DaemonException)
from blackhole.logs import configure_logs

from ._utils import (Args, cleandir, create_config, create_file, reset)


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_test():
    cfile = create_config(('', ))
    with mock.patch('sys.argv', ['blackhole', '-t', '-c', cfile]), \
        mock.patch('blackhole.config.Config.test_port', return_value=True), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '0'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_test_fails():
    cfile = create_config(('listen=127.0.0.1:0', ))
    with mock.patch('sys.argv', ['blackhole', '-t', '-c', cfile]), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_load_test_fails():
    cfile = create_config(('listen=127.0.0.1:0', ))
    with mock.patch('sys.argv', ['blackhole', '-t', '-c', cfile]), \
        mock.patch('blackhole.config.Config.test',
                   side_effect=ConfigException()), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_foreground():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile)]), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('os.chown'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers'), \
        mock.patch('blackhole.control.pid_permissions'), \
        mock.patch('blackhole.control.setgid'), \
        mock.patch('blackhole.control.setuid'), \
        mock.patch('blackhole.supervisor.Supervisor.run'), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '0'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_foreground_pid_error():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile)]), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('os.getpid', return_value=1234), \
        mock.patch('atexit.register', side_effect=DaemonException), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '64'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_foreground_socket_error():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile)]), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('blackhole.supervisor.Supervisor.close_socks'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers',
                   side_effect=BlackholeRuntimeException), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '77'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_background():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile), '-b']), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('os.chown'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers'), \
        mock.patch('blackhole.daemon.Daemon.daemonize'), \
        mock.patch('blackhole.control.pid_permissions'), \
        mock.patch('blackhole.control.setgid'), \
        mock.patch('blackhole.control.setuid'), \
        mock.patch('blackhole.supervisor.Supervisor.run'), \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '0'


@pytest.mark.usefixtures('reset', 'cleandir')
def test_run_daemon_daemonize_error():
    pidfile = os.path.join(os.getcwd(), 'blackhole-test.pid')
    cfile = create_config(('listen=127.0.0.1:9000',
                           'pidfile={}'.format(pidfile)))
    Config(cfile).load()

    with mock.patch('sys.argv', ['-c {}'.format(cfile), '-b']), \
        mock.patch('blackhole.config.Config.test'), \
        mock.patch('blackhole.config.warn_options'), \
        mock.patch('atexit.register'), \
        mock.patch('os.chown'), \
        mock.patch('blackhole.supervisor.Supervisor.generate_servers'), \
        mock.patch('os.fork', side_effect=OSError), \
        mock.patch('blackhole.supervisor.Supervisor.'
                   'close_socks') as mock_close, \
            pytest.raises(SystemExit) as err:
        run()
    assert str(err.value) == '77'
    assert mock_close.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
def test_blackhole_config():
    args = Args((('debug', False), ('quiet', False), ))
    configure_logs(args)
    mmock = mock.MagicMock(spec=logging)
    with pytest.raises(SystemExit) as err, \
            mock.patch('logging.getLogger', return_value=mmock):
        blackhole_config()
    assert str(err.value) == '2'



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import asyncio
import socket
import time
from unittest import mock

import pytest

from blackhole.child import Child
from blackhole.control import server
from blackhole.worker import Worker

from ._utils import (Args, cleandir, create_config, create_file, reset)


try:
    import asyncio
    import uvloop
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_worker_ping_pong(unused_tcp_port):
    aserver = server('127.0.0.1', unused_tcp_port, socket.AF_INET)
    started = time.monotonic()
    worker = Worker('1', [aserver, ])
    assert worker._started is True
    await asyncio.sleep(35)
    worker.stop()
    assert worker._started is False
    assert worker.ping > started
    assert worker.ping_count == 2


@pytest.mark.usefixtures('reset', 'cleandir')
@pytest.mark.asyncio
async def test_restart(unused_tcp_port):
    aserver = server('127.0.0.1', unused_tcp_port, socket.AF_INET)
    started = time.monotonic()
    worker = Worker('1', [aserver, ])
    assert worker._started is True
    await asyncio.sleep(25)
    worker.ping = time.monotonic() - 120
    old_pid = worker.pid
    await asyncio.sleep(15)
    assert worker.pid is not old_pid
    worker.stop()
    assert worker._started is False
    assert worker.ping > started
    assert worker.ping_count == 0



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Setup file."""

from setuptools import find_packages, setup

from setup_helpers import (get_version, include_file, PyTest,
                           require_python)


require_python(50725104)
__version__ = get_version('blackhole/__init__.py')

entry_points = {
    'console_scripts': (
        'blackhole = blackhole.application:run',
        'blackhole_config = blackhole.application:blackhole_config',
    )
}

tests_require = ('codecov',
                 'guzzle_sphinx_theme',
                 'isort',
                 'pycodestyle',
                 'pydocstyle==1.1.1',
                 'pyflakes',
                 'pylama',
                 'pytest',
                 'pytest-asyncio',
                 'pytest-cov',
                 'radon',
                 'sphinx', )

description = ('Blackhole is an MTA (message transfer agent) that '
               '(figuratively) pipes all mail to /dev/null.')

keywords = ('blackhole', 'mta', 'email')

classifiers = ['Development Status :: 5 - Production/Stable',
               'Environment :: Console',
               'Intended Audience :: Developers',
               'Intended Audience :: Information Technology',
               'Intended Audience :: System Administrators',
               'License :: OSI Approved :: MIT License',
               'Operating System :: POSIX :: Linux',
               'Operating System :: Unix',
               'Programming Language :: Python',
               'Programming Language :: Python :: 3.6',
               'Programming Language :: Python :: 3 :: Only',
               'Topic :: Communications :: Email',
               'Topic :: Communications :: Email :: Mail Transport Agents',
               'Topic :: Education :: Testing',
               'Topic :: Internet',
               'Topic :: Software Development',
               'Topic :: Software Development :: Testing',
               ('Topic :: Software Development :: Testing :: '
                'Traffic Generation'),
               'Topic :: System :: Networking',
               'Topic :: System :: Systems Administration',
               'Topic :: Utilities', ]

setup(name='blackhole',
      version=__version__,
      url='https://blackhole.io/',
      author='Kura',
      author_email='kura@kura.io',
      maintainer='Kura',
      maintainer_email='kura@kura.io',
      description=description,
      long_description=include_file('README.rst'),
      keywords=', '.join(keywords),
      license=include_file('LICENSE'),
      platforms=['linux'],
      packages=find_packages(exclude=["*.tests"]),
      install_requires=[],
      tests_require=tests_require,
      cmdclass={'test': PyTest},
      entry_points=entry_points,
      classifiers=classifiers,
      extras_require={
        'setproctitle': ['setproctitle', ],
        'uvloop': ['uvloop', ],
      },
      )



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import asyncio
import logging
import os
import pathlib
import tempfile

import pytest

from blackhole.utils import Singleton

logging.getLogger('blackhole').addHandler(logging.NullHandler())


@pytest.fixture()
def cleandir():
    newpath = tempfile.mkdtemp()
    os.chdir(newpath)


@pytest.fixture()
def reset():
    Singleton._instances = {}


def create_config(data):
    cwd = os.getcwd()
    path = os.path.join(pathlib.Path(cwd),
                        pathlib.Path('test.conf'))
    with open(path, 'w') as cfile:
        cfile.write('\n'.join(data))
    return path


def create_file(name, data=''):
    cwd = os.getcwd()
    path = os.path.join(pathlib.Path(cwd), pathlib.Path(name))
    with open(path, 'w') as ffile:
        ffile.write(str(data))
    return path


class Args(object):
    def __init__(self, args=None):
        if args is not None:
            for arg in args:
                setattr(self, arg[0], arg[1])



# (The MIT License)
#
# Copyright (c) 2013-2017 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# pylama:skip=1

import os
from unittest import mock

import pytest

from blackhole.daemon import Daemon, DaemonException

from ._utils import (Args, cleandir, create_config, create_file, reset)


@pytest.mark.usefixtures('reset', 'cleandir')
def test_instantiated_but_not_daemonised():
    pid = os.path.join(os.getcwd(), 'fake.pid')
    with mock.patch('os.getpid', return_value=666):
        daemon = Daemon(pid)
        assert daemon.pidfile == pid
        assert daemon.pid == 666


@pytest.mark.usefixtures('reset', 'cleandir')
def test_set_pid_invalid_path():
    with mock.patch('os.path.exists', return_value=False), \
        mock.patch('atexit.register'), \
            pytest.raises(DaemonException):
        Daemon('/fake/path.pid')


@pytest.mark.usefixtures('reset', 'cleandir')
def test_set_pid_valid_path():
    pid = os.path.join(os.getcwd(), 'fake.pid')
    with mock.patch('os.getpid', return_value=666), \
            mock.patch('atexit.register'):
        daemon = Daemon(pid)
        assert daemon.pidfile == pid
        assert daemon.pid == 666


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_pid_file_error():
    with mock.patch('os.path.exists', return_value=True):
        with mock.patch('builtins.open', side_effect=FileNotFoundError), \
            mock.patch('atexit.register'), \
                pytest.raises(DaemonException):
            Daemon('/fake/path.pid')
        with mock.patch('builtins.open', side_effect=IOError),\
            mock.patch('atexit.register'), \
                pytest.raises(DaemonException):
            Daemon('/fake/path.pid')
        with mock.patch('builtins.open', side_effect=PermissionError), \
            mock.patch('atexit.register'), \
                pytest.raises(DaemonException):
            Daemon('/fake/path.pid')
        with mock.patch('builtins.open', side_effect=OSError), \
            mock.patch('atexit.register'), \
                pytest.raises(DaemonException):
            Daemon('/fake/path.pid')


@pytest.mark.usefixtures('reset', 'cleandir')
def test_get_pid():
    pfile = create_file('test.pid', 123)
    with mock.patch('os.getpid', return_value=123), \
            mock.patch('atexit.register'):
        daemon = Daemon(pfile)
        assert daemon.pid is 123


@pytest.mark.usefixtures('reset', 'cleandir')
def test_delete_pid_no_exists():
    pfile = create_file('test.pid', 123)
    daemon = Daemon(pfile)
    with mock.patch('os.remove') as mock_rm, \
        mock.patch('atexit.register'), \
            mock.patch('os.path.exists', return_value=False):
        del daemon.pid
    assert mock_rm.called is False


@pytest.mark.usefixtures('reset', 'cleandir')
def test_delete_pid():
    pfile = create_file('test.pid', 123)
    with mock.patch('os.getpid', return_value=123), \
            mock.patch('atexit.register'):
        daemon = Daemon(pfile)
        assert daemon.pid is 123
        del daemon.pid
        assert daemon.pid is None


@pytest.mark.usefixtures('reset', 'cleandir')
def test_delete_pid_exit():
    pfile = create_file('test.pid', 123)
    with mock.patch('os.getpid', return_value=123), \
            mock.patch('atexit.register'):
        daemon = Daemon(pfile)
        assert daemon.pid is 123
        daemon._exit()
        assert daemon.pid is None


@pytest.mark.usefixtures('reset', 'cleandir')
def test_fork():
    pfile = create_file('test.pid', 123)
    with mock.patch('atexit.register'):
        daemon = Daemon(pfile)
    with mock.patch('os.fork', return_value=9999) as mock_fork, \
            mock.patch('os._exit') as mock_exit:
        daemon.fork()
    assert mock_fork.called is True
    assert mock_exit.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
def test_fork_error():
    pfile = create_file('test.pid', 123)
    with mock.patch('atexit.register'):
        daemon = Daemon(pfile)
    with mock.patch('os.fork', side_effect=OSError) as mock_fork, \
            pytest.raises(DaemonException):
        daemon.fork()
    assert mock_fork.called is True


@pytest.mark.usefixtures('reset', 'cleandir')
def test_daemonise():
    pfile = create_file('test.pid', 123)
    with mock.patch('blackhole.daemon.Daemon.fork') as mock_fork, \
        mock.patch('os.chdir') as mock_chdir, \
        mock.patch('os.setsid') as mock_setsid, \
        mock.patch('os.umask') as mock_umask, \
        mock.patch('atexit.register') as mock_atexit, \
            mock.patch('os.getpid', return_value=123):
        daemon = Daemon(pfile)
        daemon.daemonize()
    assert mock_fork.called is True
    assert mock_chdir.called is True
    assert mock_setsid.called is True
    assert mock_umask.called is True
    assert mock_atexit.called is True
